# AUTOGENERATED! DO NOT EDIT! File to edit: 08_experiment.ipynb (unless otherwise specified).

__all__ = ['get_data', 'get_optimizer', 'get_model', 'Experiment']

# Cell
from .setup import * #for GVs
from .preprocessing.vocab import * #for loading vocabs
from .preprocessing.transform import * #for loading ptlist thru EHRData
from .data import * #for EHRData
from .learn import * #for RunHistory
from .metrics import * #for auroc_score
from .models import * #for the models
from fastai.imports import *

# Cell
from addict import Dict

# Cell
def get_data(params, for_training=True):
    '''Convenience fn to get data (for training or testing) based on `data_params` in `experiment.yaml`'''
    ehr_data = EHRData(params.dataset_path, params.labels, params.age_start, params.age_stop, params.age_in_months, params.lazy_load_gpu)
    if for_training:
        return ehr_data.get_data(params.bs, params.num_workers)
    else:
        return ehr_data.get_test_data(params.bs, params.num_workers)

# Cell
def get_optimizer(model, params):
    '''Convenience fn to get optimizer based on `optim_params` in `experiment.yaml'''
    if   params.optim == 'SGD' : optimizer = optim.SGD(model.parameters(), lr=0.001)
    elif params.optim == 'Adam': optimizer = optim.Adam(model.parameters())
    else                       : optimizer = optim.Adagrad(model.parameters(), params.lr, params.lr_decay, params.weight_decay)
    return optimizer

# Cell
def get_model(dataset_path, params):
    '''Convenience fn to get model based on `model_params` in `experiment.yaml'''
    demograph_dims, rec_dims, demograph_dims_wd, rec_dims_wd = get_all_emb_dims(EhrVocabList.load(dataset_path), params.αd)

    if params.model == 'LSTM':
        model = EHR_LSTM(demograph_dims, rec_dims, demograph_dims_wd, rec_dims_wd,
                         params.lstm_layers, params.linear_layers, params.initrange, params.bn,
                         params.input_drp, params.lstm_drp, params.linear_drp, params.zero_bn)
    elif params.model == 'CNN':
        model = EHR_CNN(demograph_dims, rec_dims, demograph_dims_wd, rec_dims_wd,
                        params.linear_layers, params.initrange, params.bn,
                        params.input_drp, params.linear_drp, params.zero_bn)
    return model.to(DEVICE)

# Cell
class Experiment:
    '''A simple (bare bones) Experiment class for experiment management'''
    def __init__(self, name, config, params):
        self.name = name
        self.desc = config.desc
        self.path = Path(f'{config.path}/{name}')
        self.chkpt_path = Path(f'{config.checkpoint_path}/{name}')
        self.labels = params.data_params.labels
        self.config, self.params = config, params
        self.history = RunHistory(self.labels)

    def __repr__(self):
        '''To print out important features of an Experiment'''
        res =  f'Experiment name: {self.name}\n'
        res += f'Description: {self.desc}\n'
        res += f'Path: {self.path}\n'
        res += f'Labels: {self.params.data_params.labels}\n'
        res += f'Dataset: {self.params.data_params.dataset_path}\n'
        res += f'Model: {self.params.model_params.model}\n'
        res += f'Optim: {self.params.optim_params.optim}'
        return res

    def save(self):
        '''Save the Experiment'''
        self.path.mkdir(parents=True, exist_ok=True)
        with open(f'{self.path}/{self.name}.experiment', 'wb')  as f:
            pickle.dump(self, f)
        print(f'Saved experiment to {self.path}/{self.name}.experiment')

    @classmethod
    def load(cls, name, path=None):
        '''Load an existing Experiment'''
        exp_dir = Path(f'{EXPERIMENT_STORE}/{name}') if path == None else Path(f'{path}/{name}')
        with open(f'{exp_dir}/{name}.experiment','rb') as infile:
            exp = pickle.load(infile)
        print(f'Loaded experiment from {exp_dir}/{name}.experiment')
        return exp

    @classmethod
    def create_from_file(cls, path, name):
        '''Create a *new* Experiment object from the `experiment.yaml` file in the `path/name` directory'''
        exp_dir = Path(f'{path}/{name}')
        with open(f'{exp_dir}/experiment.yaml', 'rb') as f:
            contents = yaml.full_load(f)

        contents = Dict(contents)
        return cls(name, contents.config, contents.params)

    def fit(self, epochs, from_checkpoint=False, to_checkpoint=True,  verbosity=.75, plot=True, save=True):
        '''Fit function that assembles everything needed and calls the `lemonade.learn.fit` function'''
        train_dl, valid_dl, train_pos_wts, valid_pos_wts = get_data(self.params.data_params)
        model = get_model(self.params.data_params.dataset_path, self.params.model_params)
        train_loss_fn, valid_loss_fn = get_loss_fn(train_pos_wts), get_loss_fn(valid_pos_wts)
        optim = get_optimizer(model, self.params.optim_params)
        lazy = self.params.data_params.lazy_load_gpu
        from_chkpt_path = self.chkpt_path if from_checkpoint else None
        to_chkpt_path   = self.chkpt_path if to_checkpoint   else None

        self.history = fit(epochs, self.history, model, train_loss_fn, valid_loss_fn, optim,
                           auroc_score, train_dl, valid_dl, lazy, to_chkpt_path, from_chkpt_path, verbosity)
        if plot: plot_fit_results(self.history, self.labels)
        if save: self.save()

    def predict(self, plot=True, save=True):
        '''Predict function that assembles everything needed and calls the `lemonade.learn.predict` function'''
        test_dl, test_pos_wts = get_data(self.params.data_params, for_training=False)
        model = get_model(self.params.data_params.dataset_path, self.params.model_params)
        test_loss_fn = get_loss_fn(test_pos_wts)
        lazy = self.params.data_params.lazy_load_gpu

        self.history = predict(self.history, model, test_loss_fn, auroc_score, test_dl, self.chkpt_path, lazy)
        self.history = summarize_prediction(self.history, self.labels, plot)
        if save: self.save()


# AUTOGENERATED! DO NOT EDIT! File to edit: 06_learn.ipynb (unless otherwise specified).

__all__ = ['save_to_checkpoint', 'load_from_checkpoint', 'get_loss_fn', 'RunHistory', 'train', 'evaluate', 'fit',
           'predict', 'plot_loss', 'plot_losses', 'plot_aurocs', 'plot_train_valid_aurocs', 'plot_fit_results',
           'summarize_prediction', 'count_parameters']

# Cell
from lemonade.setup import *
from lemonade.metrics import *
from fastai.imports import *

# Cell
def save_to_checkpoint(epoch_index, model, optimizer, path):
    '''Save model and optimizer state_dicts to checkpoint'''
    if not os.path.isdir(path): Path(path).mkdir(parents=True, exist_ok=True)
    torch.save({
        'epoch_index':epoch_index,
        'model_state_dict': model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict()
        }, f'{path}/checkpoint.tar')
    print(f'Checkpointed to "{path}/checkpoint.tar"')

# Cell
def load_from_checkpoint(model, path, optimizer=None, for_inference=False):
    '''Load from checkpoint - model, optimizer & epoch_index for training or just model for inference'''

    print(f'From "{path}/checkpoint.tar", loading model ...')
    chkpt = torch.load(f'{path}/checkpoint.tar')
    model.load_state_dict(chkpt['model_state_dict'])
    model = model.to(DEVICE)

    if for_inference:
        return model
    else:
        print(f'loading optimizer and epoch_index ...')
        optimizer.load_state_dict(chkpt['optimizer_state_dict'])
        return chkpt['epoch_index'], model, optimizer

# Cell
def get_loss_fn(pos_wts):
    '''Return `nn.BCEWithLogitsLoss` with the given positive weights'''
    return nn.BCEWithLogitsLoss(pos_weight=pos_wts).to(DEVICE)

# Cell
class RunHistory:
    '''Class to hold training and prediction run histories'''
    def __init__(self, labels):
        self.train = self.valid = self.test = pd.DataFrame(columns=['loss', *labels])
        self.y_train = self.yhat_train = self.y_valid = self.yhat_valid = self.y_test = self.yhat_test = []
        self.prediction_summary = pd.DataFrame()

# Cell
def train(model, train_dl, train_loss_fn, optimizer, lazy=True):
    '''Train model using train dataset'''
    yhat_train = y_train = Tensor([])
    train_loss = 0.
    model.train()

    for xb, yb in train_dl:
        if lazy: xb, yb = [x.to_gpu(non_block=True) for x in xb], yb.to(DEVICE, non_blocking=True)
        y_hat  = model(xb)
        loss   = train_loss_fn(y_hat, yb)

        train_loss += loss.item()
        yhat_train = torch.cat((yhat_train, torch.sigmoid( y_hat.cpu().detach() ) ))
        y_train    = torch.cat((y_train, yb.cpu().detach()))

        loss.backward()
        optimizer.step()
        model.zero_grad(set_to_none=True)

    return train_loss, yhat_train, y_train, model

# Cell
def evaluate(model, eval_dl, eval_loss_fn, lazy=True):
    '''Evaluate model - used for validation (while training) and prediction'''
    yhat_eval = y_eval = Tensor([])
    eval_loss = 0.
    model.eval()

    with torch.no_grad():
        for xb, yb in eval_dl:
            if lazy: xb, yb = [x.to_gpu(non_block=True) for x in xb], yb.to(DEVICE, non_blocking=True)
            y_hat  = model(xb)

            eval_loss += (eval_loss_fn(y_hat, yb)).item()
            yhat_eval = torch.cat((yhat_eval, torch.sigmoid( y_hat.cpu().detach() ) ))
            y_eval    = torch.cat((y_eval, yb.cpu().detach()))

    return eval_loss, yhat_eval, y_eval

# Cell
def fit(epochs, history, model, train_loss_fn, valid_loss_fn, optimizer, accuracy_fn,
        train_dl, valid_dl, lazy=True, to_chkpt_path=None, from_chkpt_path=None, verbosity=0.75):
    '''Fit model and return results in `history`'''

    if from_chkpt_path:
        last_epoch, model, optimizer = load_from_checkpoint(model, from_chkpt_path, optimizer, for_inference=False)
        start_epoch = last_epoch+1
    else:
        start_epoch = 0
    end_epoch = start_epoch+(epochs-1)
    print_epochs = np.linspace(start_epoch, end_epoch, int(epochs*verbosity), endpoint=True, dtype=int)
    train_history, valid_history = [], []

    print('{:>5} {:>16} {:^20} {:>25} {:^20}'.format('epoch |', 'train loss |', 'train aurocs', 'valid loss |', 'valid aurocs'))
    print('{:-^100}'.format('-'))

    for epoch in range (start_epoch, start_epoch+epochs):

        train_loss, yhat_train, y_train, model = train(model, train_dl, train_loss_fn, optimizer, lazy)
        valid_loss, yhat_valid, y_valid = evaluate(model, valid_dl, valid_loss_fn, lazy)

        train_loss,   valid_loss   = train_loss/len(train_dl), valid_loss/len(valid_dl)
        train_aurocs, valid_aurocs = accuracy_fn(y_train, yhat_train), accuracy_fn(y_valid, yhat_valid)
        train_history.append([train_loss, *train_aurocs])
        valid_history.append([valid_loss, *valid_aurocs])

        if epoch in print_epochs:
            row  = f'{epoch:>5} |'
            row += f'{train_loss:>15.3f} | '
            row += f'{[f"{a:.3f}" for a in train_aurocs[:4]]}'
            row += f'{valid_loss:>19.3f} | '
            row += f'{[f"{a:.3f}" for a in valid_aurocs[:4]]}'
            print(re.sub("',*", "", row))

    if to_chkpt_path: save_to_checkpoint(end_epoch, model, optimizer, to_chkpt_path)

    h = history
    h.y_train, h.yhat_train, h.y_valid, h.yhat_valid = y_train, yhat_train, y_valid, yhat_valid
    h.train = h.train.append(pd.DataFrame(train_history, columns=h.train.columns), ignore_index=True)
    h.valid = h.valid.append(pd.DataFrame(valid_history, columns=h.valid.columns), ignore_index=True)

    return h

# Cell
def predict(history, model, test_loss_fn, accuracy_fn, test_dl, chkpt_path, lazy=True):
    '''Predict and return results in `history`'''
    model = load_from_checkpoint(model, chkpt_path, for_inference=True)
    test_loss, yhat_test, y_test = evaluate(model, test_dl, test_loss_fn, lazy)

    test_loss = test_loss/len(test_dl)
    test_aurocs = accuracy_fn(y_test, yhat_test)
    print(f'test loss = {test_loss}')
    print(f'test aurocs = {test_aurocs}')

    h = history
    h.test = pd.DataFrame([[test_loss, *test_aurocs]], columns=h.test.columns)
    h.y_test, h.yhat_test = y_test, yhat_test

    return h

# Cell
def plot_loss(history_df, title='Loss', axis=None):
    '''Plot loss'''
    if axis == None:
        fig = plt.figure(figsize=(8,5))
        axis = fig.add_axes([0,0,1,1])

    axis.plot(range(len(history_df)), history_df['loss'], label=title)

    axis.set_title(title)
    axis.set_xlabel('epochs')
    axis.set_ylabel('loss')
    axis.legend(loc=0)

# Cell
def plot_losses(train_history, valid_history):
    '''Plot multiple losses (train and valid) side by side'''
    fig, axes = plt.subplots(1,2, figsize=(15,5))
    plt.tight_layout()

    plot_loss(train_history, title='Train Loss', axis=axes[0])
    plot_loss(valid_history, title='Valid Loss', axis=axes[1])

# Cell
def plot_aurocs(history_df, title='AUROC Scores', axis=None):
    '''Plot AUROC scores'''
    if axis == None:
        fig = plt.figure(figsize=(8,5))
        axis = fig.add_axes([0,0,1,1])
    for lbl in history_df.columns[1:]:
        axis.plot(range(len(history_df)), history_df[lbl], label=f'{lbl} (final: {history_df[lbl].iat[-1]:.3f})')
    axis.set_title(title)
    axis.set_xlabel('epochs')
    axis.set_ylabel('auroc scores')
    axis.legend(loc=0)

# Cell
def plot_train_valid_aurocs(train_history, valid_history):
    '''Plot train and valid AUROC scores side by side'''
    fig, axes = plt.subplots(1,2, figsize=(15,5))
    plt.tight_layout()
    plot_aurocs(train_history, title='Train - AUROC Scores', axis=axes[0])
    plot_aurocs(valid_history, title='Valid - AUROC Scores', axis=axes[1])

# Cell
def plot_fit_results(history, labels):
    '''All plots after fit - ROC curves, losses and AUROCs'''
    h = history
    train_rocs, valid_rocs = MultiLabelROC(h.y_train, h.yhat_train, labels), MultiLabelROC(h.y_valid, h.yhat_valid, labels)
    plot_train_valid_rocs(train_rocs.ROCs, valid_rocs.ROCs, labels, multilabel=True)
    plot_losses(h.train, h.valid)
    plot_train_valid_aurocs(h.train, h.valid)

# Cell
def summarize_prediction(history, labels, plot=True):
    '''Summarize after prediction - plot ROC curves, calculate auroc, optimal threshold and 95% CI for AUROC and return results in `history`'''
    h = history
    test_rocs = MultiLabelROC(h.y_test, h.yhat_test, labels)
    if plot: plot_rocs(test_rocs.ROCs, labels, title='Test ROC curves', multilabel=True)
    print('\nPrediction Summary ...')
    col_names = ['auroc_score', 'optimal_threshold', 'auroc_95_ci']
    rows = []
    for i, label in enumerate(labels):
        row = [test_rocs.ROCs[label].auroc, test_rocs.ROCs[label].optimal_thresh(), auroc_ci(h.y_test[:,i], h.yhat_test[:,i])]
        rows.append(row)
    history.prediction_summary = pd.DataFrame(rows, index=labels, columns=col_names)
    print(history.prediction_summary)
    return history

# Cell
def count_parameters(model, printout=False):
    '''Returns number of parameters in model'''
    total         = sum(p.numel() for p in model.parameters())
    trainable     = sum(p.numel() for p in model.parameters() if p.requires_grad)
    non_trainable = sum(p.numel() for p in model.parameters() if p.requires_grad==False)
    assert total == (trainable+non_trainable)
    print(f'total: {total:,}, trainable: {trainable:,}, non_trainable: {non_trainable:,}')
    return total, trainable, non_trainable


# AUTOGENERATED! DO NOT EDIT! File to edit: 07_models.ipynb (unless otherwise specified).

__all__ = ['dropout_mask', 'InputDropout', 'linear_layer', 'create_linear_layers', 'init_lstm', 'EHR_LSTM', 'init_cnn',
           'conv_layer', 'EHR_CNN']

# Cell
from lemonade.setup import *
from lemonade.preprocessing.vocab import * #for loading vocabs
from lemonade.preprocessing.transform import * #for loading ptlist thru EHRData
from lemonade.data import * #for EHRData
from lemonade.learn import * #for fit/predict stuff
from lemonade.metrics import * #for auroc_score
from fastai.imports import *

# Cell
def dropout_mask(x, sz, p):
    '''Dropout mask as described in fast.ai'''
    return x.new(*sz).bernoulli_(1-p).div_(1-p)

class InputDropout(nn.Module):
    '''InputDropout - same as RNNDropout described in fast.ai'''
    def __init__(self, p=0.5):
        super().__init__()
        self.p=p

    def forward(self, x):
        if not self.training or self.p == 0.: return x
        m = dropout_mask(x.data, (x.size(0), 1, x.size(2)), self.p)
        return x * m

# Cell
def linear_layer(in_features, out_features, bn=False, dropout_p=0.0):
    '''Create a single linear layer'''
    layer = [nn.Linear(in_features, out_features)]
    if bn: layer.append(nn.BatchNorm1d(out_features))
    layer.append(nn.ReLU(inplace=True))
    layer.append(nn.Dropout(dropout_p))
    return layer

def create_linear_layers(in_features_start, num_layers, bn=False, dropout_p=0.0):
    '''Create linear layers'''
    layers = []

    for l in range(num_layers):
        in_features = in_features_start if l==0 else out_features
        out_features = in_features * 2
        layers.extend(linear_layer(in_features, out_features, bn, dropout_p))

    return out_features, nn.Sequential(*layers)

# Cell
def init_lstm(m, initrange, zero_bn=False):
    '''Initialize LSTM'''
    if isinstance(m, (nn.Embedding, nn.EmbeddingBag)): m.weight.data.uniform_(-initrange, initrange)
    if isinstance(m, (nn.LSTM, nn.Linear)):
        for name, param in m.named_parameters():
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.kaiming_normal_(param)
    if isinstance(m, (nn.BatchNorm1d)): nn.init.constant_(m.weight, 0. if zero_bn else 1.)
    for l in m.children(): init_lstm(l, initrange, zero_bn)

# Cell
from fastai.layers import BatchNorm1dFlat
class EHR_LSTM(nn.Module):
    '''Based on LSTM described in this paper - https://arxiv.org/abs/1801.07860'''

    def __init__(self, demograph_dims, rec_dims, demograph_wd, rec_wd, lstm_layers=4, linear_layers=4,
                 initrange=0.3, bn=False, input_drp=0.3, lstm_drp=0.3, linear_drp=0.3, zero_bn=False):

        super().__init__()

        self.embs  = nn.ModuleList([nn.Embedding(*dim) for dim in demograph_dims])
        self.embgs = nn.ModuleList([nn.EmbeddingBag(*dim) for dim in rec_dims])

        self.rec_wd        = rec_wd
        self.demograph_wd  = demograph_wd
        self.nh            = rec_wd
        self.lstm_layers   = lstm_layers
        self.bn            = bn
        lin_features_start = (demograph_wd + 1) + rec_wd #adding 1 for age_now


        self.input_dp = InputDropout(input_drp)
        self.lstm     = nn.LSTM(input_size=self.nh, hidden_size=self.nh, num_layers=lstm_layers, batch_first=True, dropout=lstm_drp)
        if bn: self.lstm_bn = BatchNorm1dFlat(self.nh) #fastai - `nn.BatchNorm1d`, but first flattens leading dimensions
        out, self.lin = create_linear_layers(lin_features_start, linear_layers, bn, linear_drp)
        self.lin_o    = nn.Linear(out, 4)

        init_lstm(self, initrange, zero_bn)


    def get_embs(self, ptbatch_recs, ptbatch_demogs, x):
        for p in range(len(x)): #for the batch of pts
            ptbatch_recs[p] = torch.cat((self.embgs[0](x[p].obs_nums,x[p].obs_offsts),
                                         self.embgs[1](x[p].alg_nums,x[p].alg_offsts),
                                         self.embgs[2](x[p].crpl_nums,x[p].crpl_offsts),
                                         self.embgs[3](x[p].med_nums,x[p].med_offsts),
                                         self.embgs[4](x[p].img_nums,x[p].img_offsts),
                                         self.embgs[5](x[p].proc_nums,x[p].proc_offsts),
                                         self.embgs[6](x[p].cnd_nums,x[p].cnd_offsts),
                                         self.embgs[7](x[p].imm_nums,x[p].imm_offsts)),
                                        dim=1) #for the entire age span, example all 20 yrs
            ptbatch_demogs[p] = torch.cat((self.embs[0](x[p].demographics[0]),
                                           self.embs[1](x[p].demographics[1]),
                                           self.embs[2](x[p].demographics[2]),
                                           self.embs[3](x[p].demographics[3]),
                                           self.embs[4](x[p].demographics[4]),
                                           self.embs[5](x[p].demographics[5]),
                                           self.embs[6](x[p].demographics[6]),
                                           self.embs[7](x[p].demographics[7]),
                                           self.embs[8](x[p].demographics[8]),
                                           self.embs[9](x[p].demographics[9]),
                                           self.embs[10](x[p].demographics[10]),
                                           x[p].age_now))

        return ptbatch_recs, ptbatch_demogs

    def forward(self, x):

        bs   = len(x)
        bptt = len(x[0].obs_offsts)
        h    = torch.zeros(self.lstm_layers,bs,self.nh, device=DEVICE)

        ptbatch_recs   = torch.empty(bs,bptt,self.rec_wd, device=DEVICE)
        ptbatch_demogs = torch.empty(bs,self.demograph_wd+1, device=DEVICE)
        ptbatch_recs, ptbatch_demogs = self.get_embs(ptbatch_recs, ptbatch_demogs, x)

        ptbatch_recs = self.input_dp(ptbatch_recs)                                     #apply input dropout

        res,_ = self.lstm(ptbatch_recs, (h,h))
        res   = self.lstm_bn(res[:,-1]) if self.bn else res[:,-1]                      #lstm output
        res   = self.lin(torch.cat((res, ptbatch_demogs), dim=1))                      #concat demographics + send thru linear lyrs
        out   = self.lin_o(res)

        return out

# Cell
def init_cnn(m, initrange, zero_bn=False):
    '''Initialize CNN as described in fast.ai'''
    if getattr(m, 'bias', None) is not None: nn.init.constant_(m.bias, 0)
    if isinstance(m, (nn.Embedding, nn.EmbeddingBag)): m.weight.data.uniform_(-initrange, initrange)
    if isinstance(m, (nn.Conv2d, nn.Linear)): nn.init.kaiming_normal_(m.weight)
    if isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d)): nn.init.constant_(m.weight, 0. if zero_bn else 1.)
    for l in m.children(): init_cnn(l, initrange, zero_bn)

def conv_layer(in_channels, out_channels, kernel_size, stride=1, padding=1, bn=False):
    '''Create a single conv layer - as described in fast.ai'''
    layer = [nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, bias=not bn)]
    if bn: layer.append(nn.BatchNorm2d(out_channels))
    layer.append(nn.ReLU(inplace=True))
    return layer

# Cell
class EHR_CNN(nn.Module):
    '''Based on the model described in the Deepr paper - https://arxiv.org/abs/1607.07519'''
    def __init__(self, demograph_dims, rec_dims, demograph_wd, rec_wd, linear_layers=4,
                 initrange=0.3, bn=False, input_drp=0.3, linear_drp=0.3, zero_bn=False):

        super().__init__()

        self.embs  = nn.ModuleList([nn.Embedding(*dim) for dim in demograph_dims])
        self.embgs = nn.ModuleList([nn.EmbeddingBag(*dim) for dim in rec_dims])

        self.rec_wd        = rec_wd
        self.demograph_wd  = demograph_wd
        self.bn            = bn
        lin_features_start = (demograph_wd + 1) + 512 #adding 1 for 'age_now', 512 = cnn output


        self.input_dp = InputDropout(input_drp)

        out, self.lin = create_linear_layers(lin_features_start, linear_layers, bn, linear_drp)
        self.lin_o    = nn.Linear(out, 4)

        self.cnn = nn.Sequential(
        *conv_layer(in_channels=1, out_channels=2, kernel_size=(5,5), padding=2, bn=self.bn),
        *conv_layer(2,4,kernel_size=(3,3), padding=1, bn=self.bn),
        *conv_layer(4,8,kernel_size=3, stride=2, padding=1, bn=self.bn),
        *conv_layer(8,16,kernel_size=3, stride=2, padding=1, bn=self.bn),
        *conv_layer(16,32,kernel_size=3, stride=2, padding=1, bn=self.bn),
        nn.AdaptiveMaxPool2d((4,4)),
        nn.Flatten()
        )

        init_cnn(self, initrange, zero_bn)


    def get_embs(self, ptbatch_recs, ptbatch_demogs, x):
        for p in range(len(x)): #for the batch of pts
            ptbatch_recs[p] = torch.cat((self.embgs[0](x[p].obs_nums,x[p].obs_offsts),
                                         self.embgs[1](x[p].alg_nums,x[p].alg_offsts),
                                         self.embgs[2](x[p].crpl_nums,x[p].crpl_offsts),
                                         self.embgs[3](x[p].med_nums,x[p].med_offsts),
                                         self.embgs[4](x[p].img_nums,x[p].img_offsts),
                                         self.embgs[5](x[p].proc_nums,x[p].proc_offsts),
                                         self.embgs[6](x[p].cnd_nums,x[p].cnd_offsts),
                                         self.embgs[7](x[p].imm_nums,x[p].imm_offsts)),
                                        dim=1) #for the entire age span, example all 20 yrs
            ptbatch_demogs[p] = torch.cat((self.embs[0](x[p].demographics[0]),
                                           self.embs[1](x[p].demographics[1]),
                                           self.embs[2](x[p].demographics[2]),
                                           self.embs[3](x[p].demographics[3]),
                                           self.embs[4](x[p].demographics[4]),
                                           self.embs[5](x[p].demographics[5]),
                                           self.embs[6](x[p].demographics[6]),
                                           self.embs[7](x[p].demographics[7]),
                                           self.embs[8](x[p].demographics[8]),
                                           self.embs[9](x[p].demographics[9]),
                                           self.embs[10](x[p].demographics[10]),
                                           x[p].age_now))

        return ptbatch_recs, ptbatch_demogs

    def forward(self, x):

        bs     = len(x)
        height = len(x[0].obs_offsts)
        width  = self.rec_wd

        ptbatch_recs   = torch.empty(bs,height,width, device=DEVICE)
        ptbatch_demogs = torch.empty(bs,self.demograph_wd+1, device=DEVICE)

        ptbatch_recs, ptbatch_demogs = self.get_embs(ptbatch_recs, ptbatch_demogs, x)
        ptbatch_recs = self.input_dp(ptbatch_recs)                                    #apply input dropout

        res = self.cnn(ptbatch_recs.reshape(bs,1,height,width))                       #cnn output
        res   = self.lin(torch.cat((res, ptbatch_demogs), dim=1))                     #concat demographics + send thru linear lyrs
        out   = self.lin_o(res)

        return out


# AUTOGENERATED! DO NOT EDIT! File to edit: 04_data.ipynb (unless otherwise specified).

__all__ = ['EHRDataSplits', 'LabelEHRData', 'EHRDataset', 'EHRData']

# Cell
from .setup import *
from .preprocessing.transform import *
from fastai.imports import *
import copy

# Cell
class EHRDataSplits():
    '''Class to hold the PatientList splits; defaults to loading 0 to 20 years age span'''
    def __init__(self, path, age_start=0, age_stop=20, age_in_months=False):
        self.train, self.valid, self.test = self._load_splits(path, age_start, age_stop, age_in_months)

    def _load_splits(self, path, age_start, age_stop, age_in_months):
        '''Load splits of preprocessed `PatientList`s from persistent store using path'''
        train = PatientList.load(path, 'train', age_start, age_stop, age_in_months)
        valid = PatientList.load(path, 'valid', age_start, age_stop, age_in_months)
        test  = PatientList.load(path, 'test',  age_start, age_stop, age_in_months)
        return train, valid, test

    def get_splits(self):
        '''Return splits'''
        return self.train, self.valid, self.test

    def get_lengths(self):
        '''Return a dataframe with lengths (# of patients) of the splits (train, valid, test) and total'''
        lengths = [len(self.train), len(self.valid), len(self.test), len(self.train)+len(self.valid)+len(self.test)]
        return pd.DataFrame(lengths, index=['train','valid','test','total'], columns=['lengths'])

    def get_label_counts(self, labels):
        '''Get prevalence counts of labels in each split - returns a dataframe with counts for each split and total count'''
        counts = []
        for label in labels:
            train_count = [getattr(self.train[i],label) == 1 for i in range(len(self.train))].count(True)
            valid_count = [getattr(self.valid[i],label) == 1 for i in range(len(self.valid))].count(True)
            test_count  = [getattr(self.test[i],label) == 1 for i in range(len(self.test))].count(True)
            total_count = train_count+valid_count+test_count
            counts.append([train_count, valid_count, test_count, total_count])
        return pd.DataFrame(counts, index=labels, columns=['train','valid','test','total'])

    def get_pos_wts(self, labels):
        '''Get positive weights to be used in `nn.BCEWithLogitsLoss`'''
        pos_counts = self.get_label_counts(labels)
        neg_counts = self.get_lengths().transpose().values - pos_counts
        return round(neg_counts / pos_counts)

# Cell
class LabelEHRData():
    '''Class to hold labeled EHR data splits'''
    def __init__(self, train, valid, test, labels):
        '''Extracts y from patient object, each labelset a tuple of x,y: x=Patient object, y=tensor of conditions'''
        self.x_train, self.y_train = train, self._get_y(train, labels)
        self.x_valid, self.y_valid = valid, self._get_y(valid, labels)
        self.x_test,  self.y_test  = test , self._get_y(test , labels)

        self.train = self.x_train, self.y_train
        self.valid = self.x_valid, self.y_valid
        self.test  = self.x_test,  self.y_test

    def _get_y(self, ds, labels):
        '''Extract y from each patient object in ds and stack them - ds is dataset containing patient objects'''
        y = []
        for pt in ds:
            y.append( torch.FloatTensor(np.array([getattr(pt,label) for label in labels], dtype='float')) )
        return torch.stack(y)

# Cell
class EHRDataset(torch.utils.data.Dataset):
    '''Class to hold a single EHR dataset (holds a tuple of x and y) -- handles lazy vs full loading of dataset on GPU'''
    def __init__(self, x_labeled, y_labeled, lazy_load_gpu=True):
        '''If `lazy_load_gpu` is `False`, load entire dataset on GPU'''
        if lazy_load_gpu:
            self.x, self.y = x_labeled, y_labeled
            self.lazy = True
        else:
            self.x, self.y = [x.to_gpu() for x in x_labeled], y_labeled.to(DEVICE)
            self.lazy = False

    def __len__(self): return len(self.x)

    def _test_getitem(self, i): return self.x[i],self.y[i]

    def __getitem__(self, i):
        '''If lazy loading, return deep copy of patient object `i`, else entire dataset already on GPU - just return `i`'''
        if self.lazy:
            return copy.deepcopy(self.x[i]), self.y[i]
        else:
            return self.x[i], self.y[i]

# Cell
class EHRData:
    '''All encompassing class for EHR data - holds Splits, Labels, Datasets, DataLoaders and provides convenience fns for training and prediction'''
    def __init__(self, path, labels, age_start=0, age_stop=20, age_in_months=False, lazy_load_gpu=True):
        self.path, self.labels = path, labels
        self.age_start, self.age_stop, self.age_in_months = age_start, age_stop, age_in_months
        self.lazy_load_gpu = lazy_load_gpu

    def load_splits(self):
        '''Load data splits given dataset path'''
        self.splits = EHRDataSplits(self.path, self.age_start, self.age_stop, self.age_in_months)

    def label(self):
        '''Run labeler - i.e. extract y from patient objects'''
        self.labeled = LabelEHRData(*self.splits.get_splits(), self.labels)

    def create_datasets(self):
        '''Create `EHRDataset`s'''
        self.train_ds = EHRDataset(*self.labeled.train, self.lazy_load_gpu)
        self.valid_ds = EHRDataset(*self.labeled.valid, self.lazy_load_gpu)
        self.test_ds  = EHRDataset(*self.labeled.test, self.lazy_load_gpu)

    def ehr_collate(b):
        '''Custom collate function for use in `DataLoader`'''
        xs,ys = zip(*b)
        return xs, torch.stack(ys)

    def create_dls(self, bs, lazy, c_fn=ehr_collate, **kwargs):
        '''Create `DataLoader`s'''
        self.train_dl = DataLoader(self.train_ds, bs, shuffle=True, collate_fn=c_fn, pin_memory=lazy, **kwargs)
        self.valid_dl = DataLoader(self.valid_ds, bs*2, collate_fn=c_fn, pin_memory=lazy, **kwargs)
        self.test_dl  = DataLoader(self.test_ds,  bs*2, collate_fn=c_fn, pin_memory=lazy, **kwargs)

    def get_data(self, bs=64, num_workers=0):
        '''Convenience function - returns everything needed for training'''
        self.load_splits()
        self.label()
        self.create_datasets()
        self.create_dls(bs, self.lazy_load_gpu, num_workers=num_workers)

        pos_wts = self.splits.get_pos_wts(self.labels)
        train_pos_wts = torch.Tensor(pos_wts['train'].values)
        valid_pos_wts = torch.Tensor(pos_wts['valid'].values)
        return self.train_dl, self.valid_dl, train_pos_wts, valid_pos_wts

    def get_test_data(self, bs=64, num_workers=0):
        '''Convenience function - returns everything needed for prediction using test data'''
        self.load_splits()
        self.label()
        self.create_datasets()
        self.create_dls(bs, self.lazy_load_gpu, num_workers=num_workers)

        pos_wts = self.splits.get_pos_wts(self.labels)
        test_pos_wts = torch.Tensor(pos_wts['test'].values)
        return self.test_dl, test_pos_wts


# AUTOGENERATED! DO NOT EDIT! File to edit: 00_basics.ipynb (unless otherwise specified).

__all__ = ['get_device', 'read_settings', 'DEVICE', 'settings', 'DATA_STORE', 'LOG_STORE', 'MODEL_STORE',
           'EXPERIMENT_STORE', 'PATH_1K', 'PATH_10K', 'PATH_20K', 'PATH_100K', 'FILENAMES', 'SYNTHEA_DATAGEN_DATES',
           'CONDITIONS', 'LABELS', 'LOG_NUMERICALIZE_EXCEP']

# Cell
from fastai.imports import *
from addict import Dict

# Cell
def get_device():
    '''Checks to see if GPU is available and sets device to GPU or CPU'''
    use_cuda = torch.cuda.is_available()
    if use_cuda:
        assert torch.backends.cudnn.enabled == True
        torch.backends.cudnn.benchmark = True #Enable cuDNN auto-tuner - perf benefit for convs
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    return device

# Cell
def read_settings():
    '''Read settings file at "~/.lemonpie/settings.yaml", if doesnt exist, create it from template'''
    settings_dir = f'{Path.home()}/.lemonpie'
    settings_file = Path(f'{settings_dir}/settings.yaml')

    if not settings_file.exists():
        print('No settings file found, so creating from template ..')
        with open('./templates/settings_template.yaml', 'r') as t:
            template = Dict(yaml.full_load(t))
        template.STORES.DATA_STORE       = f'{Path.home()}/.lemonpie/datasets'
        template.STORES.LOG_STORE        = f'{Path.home()}/.lemonpie/logs'
        template.STORES.MODEL_STORE      = f'{Path.home()}/.lemonpie/models'
        template.STORES.EXPERIMENT_STORE = f'{Path.home()}/.lemonpie/experiments'

        settings = template
        Path.mkdir(Path(settings_dir), exist_ok=True)
        with open(settings_file, 'w') as s:
            yaml.dump(settings.to_dict(), s)
    else:
        with open(settings_file, 'r') as s:
            settings = Dict(yaml.full_load(s))

    return settings

# Cell
DEVICE = get_device()
settings = read_settings()

DATA_STORE         = settings.STORES.DATA_STORE
LOG_STORE          = settings.STORES.LOG_STORE
MODEL_STORE        = settings.STORES.MODEL_STORE
EXPERIMENT_STORE   = settings.STORES.EXPERIMENT_STORE

PATH_1K   = f'{DATA_STORE}/synthea/1K'
PATH_10K  = f'{DATA_STORE}/synthea/10K'
PATH_20K  = f'{DATA_STORE}/synthea/20K'
PATH_100K = f'{DATA_STORE}/synthea/100K'

FILENAMES = settings.FILENAMES

SYNTHEA_DATAGEN_DATES = settings.SYNTHEA_DATAGEN_DATES

CONDITIONS = settings.CONDITIONS

LABELS = settings.LABELS

LOG_NUMERICALIZE_EXCEP = settings.LOG_NUMERICALIZE_EXCEP


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_preprocessing_clean.ipynb (unless otherwise specified).

__all__ = ['read_raw_ehrdata', 'split_patients', 'split_ehr_dataset', 'cleanup_pts', 'cleanup_obs', 'cleanup_algs',
           'cleanup_crpls', 'cleanup_meds', 'cleanup_img', 'cleanup_procs', 'cleanup_cnds', 'cleanup_immns',
           'cleanup_dataset', 'extract_ys', 'insert_age', 'clean_raw_ehrdata', 'load_cleaned_ehrdata',
           'load_ehr_vocabcodes']

# Cell
from ..setup import *
from fastai.imports import *

# Cell
def read_raw_ehrdata(path, csv_names = FILENAMES):
    '''Read raw EHR data'''
    dfs = [pd.read_csv(f'{path}/{fname}.csv', low_memory=False) for fname in csv_names]
    return dfs

# Cell
def split_patients(patients, valid_pct=0.2, test_pct=0.2, random_state=1234):
    '''Split the patients dataframe'''
    train_pct = 1 - (valid_pct + test_pct)
    print(f'Splits:: train: {train_pct}, valid: {valid_pct}, test: {test_pct}')
    patients = patients.sample(frac=1, random_state=random_state).reset_index(drop=True)
    return np.split(patients, [int(train_pct*len(patients)), int((train_pct+valid_pct)*len(patients))])

# Cell
def split_ehr_dataset(path, valid_pct=0.2, test_pct=0.2, random_state=1234):
    '''Split EHR dataset into train, valid, test and save'''

    train_dfs, valid_dfs, test_dfs = [],[],[]

    dfs = read_raw_ehrdata(f'{path}/raw_original')
    all_pts = dfs[0]
    all_pts.rename(str.lower, axis='columns', inplace=True)
    train_pt, valid_pt, test_pt = split_patients(dfs[0], valid_pct, test_pct, random_state)
    train_dfs.append(train_pt)
    valid_dfs.append(valid_pt)
    test_dfs.append(test_pt)
    print(f'Split {FILENAMES[0]} into:: Train: {len(train_pt)}, Valid: {len(valid_pt)}, Test: {len(test_pt)} -- Total before split: {len(dfs[0])}')

    for df, name in zip(dfs[1:], FILENAMES[1:]):
        df = df.set_index('PATIENT')
        df_train = df.loc[df.index.intersection(train_pt['id']).unique()]
        df_valid = df.loc[df.index.intersection(valid_pt['id']).unique()]
        df_test = df.loc[df.index.intersection(test_pt['id']).unique()]
        assert len(df) == len(df_train)+len(df_valid)+len(df_test),f'Split failed {name}: {len(df)} != {len(df_train)}+{len(df_valid)}+{len(df_test)}'
        train_dfs.append(df_train.reset_index())
        valid_dfs.append(df_valid.reset_index())
        test_dfs.append(df_test.reset_index())


    for split in ['train', 'valid', 'test']:
        d = Path(f'{path}/raw_split/{split}')
        d.mkdir(parents=True, exist_ok=True)

        if split == 'train':
            for df, name in zip(train_dfs, FILENAMES):
                df.to_csv(f'{d}/{name}.csv', index=False)
            print(f'Saved train data to {d}')

        if split == 'valid':
            for df, name in zip(valid_dfs, FILENAMES):
                df.to_csv(f'{d}/{name}.csv', index=False)
            print(f'Saved valid data to {d}')

        if split == 'test':
            for df, name in zip(test_dfs, FILENAMES):
                df.to_csv(f'{d}/{name}.csv', index=False)
            print(f'Saved test data to {d}')

# Cell
def cleanup_pts(pts, is_train, today=None):
    '''Clean patients df'''

    pts.rename(str.lower, axis='columns', inplace=True)
    pts = pts.loc[:, ['id', 'birthdate', 'marital', 'race', 'ethnicity', 'gender', 'birthplace', 'city', 'state', 'zip']]
    pts.rename(columns={"id":"patient"}, inplace=True)
    pts = pts.astype({'birthdate':'datetime64'})
    pts['zip'] = pts['zip'].fillna(0.0).astype(int)
    if today == None: today = pd.Timestamp.today()
    else            : today = pd.to_datetime(today)
    pts['age_now_days'] = pts['birthdate'].apply(lambda bday: (today-bday).days)

    pts.fillna('xxxnan', inplace=True)
    if is_train: pt_codes = pts.drop(columns=['patient'], inplace=False)
    pts.set_index('patient', inplace=True)
    pt_demographics = pts
    patients = pts.loc[:,['birthdate']]

    return [patients, pt_demographics, pt_codes] if is_train else [patients, pt_demographics]

# Cell
def cleanup_obs(obs, is_train):
    '''Clean observations df'''

    obs.rename(str.lower, axis='columns', inplace=True)
    obs.units.fillna('xxxnan', inplace=True)
    obs.dropna(subset=['value'], inplace=True)

    obs.rename(columns={"code":"orig_code", "description":"desc"}, inplace=True)
    obs['code'] = obs['orig_code'].str.cat(obs[['value', 'units', 'type']].astype(str), sep='||')

    if is_train: obs_codes = obs.loc[:, ['orig_code', 'desc', 'value', 'units', 'type']]

    obs = obs.loc[:, ['patient', 'date', 'code']]
    obs = obs.astype({'date':'datetime64'})
    obs.set_index('patient', inplace=True)

    return [obs, obs_codes] if is_train else [obs]

# Cell
def cleanup_algs(allergies, is_train):
    '''Clean allergies df'''

    allergies.rename(str.lower, axis='columns', inplace=True)
    allergies.drop(columns=['encounter'], inplace=True)

    stops = pd.DataFrame(allergies.loc[allergies['stop'].notnull(),:])
    allergies['code'] = allergies['code'].apply(lambda x: f'{str(x)}||START')
    stops['code'] = stops['code'].apply(lambda x: f'{str(x)}||STOP')
    allergies.drop(columns=['stop'], inplace=True)
    stops.drop(columns=['start'], inplace=True)
    allergies.rename(columns={"start":"date", "description":"desc"}, inplace=True)
    stops.rename(columns={"stop":"date", "description":"desc"}, inplace=True)
    allergies = allergies.append(stops, ignore_index=True)

    if is_train: alg_codes = allergies.loc[:, ['code', 'desc']]

    allergies.drop(columns=['desc'], inplace=True)
    allergies = allergies.astype({'date':'datetime64'})
    allergies.set_index('patient', inplace=True)
    return [allergies, alg_codes] if is_train else [allergies]

# Cell
def cleanup_crpls(careplans, is_train):
    '''Clean careplans df'''

    careplans.rename(str.lower, axis='columns', inplace=True)
    careplans = careplans.loc[:, ['start', 'stop', 'patient', 'code', 'description']]

    stops = pd.DataFrame(careplans.loc[careplans['stop'].notnull(),:])
    careplans['code'] = careplans['code'].apply(lambda x: f'{str(x)}||START')
    stops['code'] = stops['code'].apply(lambda x: f'{str(x)}||STOP')
    careplans.drop(columns=['stop'], inplace=True)
    stops.drop(columns=['start'], inplace=True)
    careplans.rename(columns={"start":"date", "description":"desc"}, inplace=True)
    stops.rename(columns={"stop":"date", "description":"desc"}, inplace=True)
    careplans = careplans.append(stops, ignore_index=True)

    if is_train: crpl_codes = careplans.loc[:, ['code', 'desc']]

    careplans.drop(columns=['desc'], inplace=True)
    careplans = careplans.astype({'date':'datetime64'})
    careplans.set_index('patient', inplace=True)
    return [careplans, crpl_codes] if is_train else [careplans]

# Cell
def cleanup_meds(medications, is_train):
    '''Clean `medications` df'''

    medications.rename(str.lower, axis='columns', inplace=True)
    medications = medications.loc[:, ['start', 'stop', 'patient', 'code', 'description']]

    stops = pd.DataFrame(medications.loc[medications['stop'].notnull(),:])
    medications['code'] = medications['code'].apply(lambda x: f'{str(x)}||START')
    stops['code'] = stops['code'].apply(lambda x: f'{str(x)}||STOP')
    medications.drop(columns=['stop'], inplace=True)
    stops.drop(columns=['start'], inplace=True)
    medications.rename(columns={"start":"date", "description":"desc"}, inplace=True)
    stops.rename(columns={"stop":"date", "description":"desc"}, inplace=True)
    medications = medications.append(stops, ignore_index=True)

    if is_train: med_codes = medications.loc[:, ['code', 'desc']]

    medications.drop(columns=['desc'], inplace=True)
    medications = medications.astype({'date':'datetime64'})
    medications.set_index('patient', inplace=True)
    return [medications, med_codes] if is_train else [medications]

# Cell
def cleanup_img(imaging_studies, is_train):
    '''Clean `imaging` df'''

    imaging_studies.rename(str.lower, axis='columns', inplace=True)
    imaging_studies.rename(columns={"bodysite_code":"code", "bodysite_description":"desc"}, inplace=True)
    if is_train: img_codes = imaging_studies.loc[:, ['code', 'desc']]

    imaging_studies = imaging_studies.loc[:, ['patient', 'date', 'code']]
    imaging_studies = imaging_studies.astype({'date':'datetime64'})
    imaging_studies.set_index('patient', inplace=True)
    return [imaging_studies, img_codes] if is_train else [imaging_studies]

# Cell
def cleanup_procs(procedures, is_train):
    '''Clean `procedures` df'''

    procedures.rename(str.lower, axis='columns', inplace=True)
    procedures.rename(columns={"description":"desc"}, inplace=True)
    if is_train: proc_codes = procedures.loc[:, ['code', 'desc']]

    procedures = procedures.loc[:, ['patient', 'date', 'code']]
    procedures = procedures.astype({'date':'datetime64'})
    procedures.set_index('patient', inplace=True)
    return [procedures, proc_codes] if is_train else [procedures]

# Cell
def cleanup_cnds(conditions, is_train):
    '''Clean `conditions` df'''

    conditions.rename(str.lower, axis='columns', inplace=True)
    conditions.drop(columns=['encounter'], inplace=True)
    stops = pd.DataFrame(conditions.loc[conditions['stop'].notnull(),:])
    conditions['code'] = conditions['code'].apply(lambda x: f'{str(x)}||START')
    stops['code'] = stops['code'].apply(lambda x: f'{str(x)}||STOP')
    conditions.drop(columns=['stop'], inplace=True)
    stops.drop(columns=['start'], inplace=True)
    conditions.rename(columns={"start":"date", "description":"desc"}, inplace=True)
    stops.rename(columns={"stop":"date","description":"desc"}, inplace=True)
    conditions = conditions.append(stops, ignore_index=True)

    if is_train: cnd_codes = conditions.loc[:, ['code', 'desc']]

    conditions.drop(columns=['desc'], inplace=True)
    conditions = conditions.astype({'date':'datetime64'})
    conditions.set_index('patient', inplace=True)
    return [conditions, cnd_codes] if is_train else [conditions]

# Cell
def cleanup_immns(immunizations, is_train):
    '''Clean `immunizations` df'''

    immunizations.rename(str.lower, axis='columns', inplace=True)
    immunizations.rename(columns={"description":"desc"}, inplace=True)
    if is_train: imm_codes = immunizations.loc[:, ['code', 'desc']]

    immunizations = immunizations.loc[:, ['patient', 'date', 'code']]
    immunizations = immunizations.astype({'date':'datetime64'})
    immunizations.set_index('patient', inplace=True)
    return [immunizations, imm_codes] if is_train else [immunizations]

# Cell
def cleanup_dataset(path, is_train, today=None):
    '''Clean all dfs in a split'''
    dfs = read_raw_ehrdata(path)

    pt_data   = cleanup_pts(dfs[0],   is_train, today)
    obs_data  = cleanup_obs(dfs[1],   is_train)
    alg_data  = cleanup_algs(dfs[2],  is_train)
    crpl_data = cleanup_crpls(dfs[3], is_train)
    med_data  = cleanup_meds(dfs[4],  is_train)
    img_data  = cleanup_img(dfs[5],   is_train)
    proc_data = cleanup_procs(dfs[6], is_train)
    cnd_data  = cleanup_cnds(dfs[7],  is_train)
    imm_data  = cleanup_immns(dfs[8], is_train)

    data_tables = [pt_data[0], pt_data[1], obs_data[0], alg_data[0], crpl_data[0], med_data[0], img_data[0], proc_data[0], cnd_data[0], imm_data[0]]
    if is_train:
        code_tables = [pt_data[2], obs_data[1], alg_data[1], crpl_data[1], med_data[1], img_data[1], proc_data[1], cnd_data[1], imm_data[1]]

    return (data_tables, code_tables) if is_train else (data_tables)

# Cell
def extract_ys(patients, conditions, cnd_dict=CONDITIONS):
    '''Extract labels from conditions df and add them to patients df with age'''
    for key in cnd_dict.keys():
        patients = patients.merge(conditions[conditions.code==cnd_dict[key]], how='left', left_index=True, right_index=True)
        patients[f'{key}_y'] = patients.code.notna()
        patients[f'{key}_age'] = ((patients.date - patients.birthdate)//np.timedelta64(1,'Y'))
        patients = patients.drop(columns=['date','code'])
    return patients

# Cell
def insert_age(df, pts_df):
    '''Insert age in years and months into each of the rec dfs'''
    df = df.merge(pts_df, left_index=True, right_index=True)
    df['age']        = (df['date'] - df['birthdate'])//np.timedelta64(1,'Y')
    df['age_months'] = (df['date'] - df['birthdate'])//np.timedelta64(1,'M')
    return df.drop(columns=['date','birthdate'])

# Cell
def clean_raw_ehrdata(path, valid_pct=0.2, test_pct=0.2, today=None):
    '''Split, clean, preprocess & save raw EHR data'''
    split_ehr_dataset(path, valid_pct, test_pct)

    for split in ['train', 'valid', 'test']:
        split_path = f'{path}/raw_split/{split}'
        if split == 'train': data_tables, code_tables = cleanup_dataset(split_path, is_train=True, today=today)
        else               : data_tables = cleanup_dataset(split_path, is_train=False)
        patients, conditions, rec_tables = data_tables[0], data_tables[8], data_tables[2:]
        patients = extract_ys(patients, conditions, CONDITIONS)
        rec_dfs = [insert_age(rec_df, pd.DataFrame(patients.birthdate)) for rec_df in rec_tables]

        cleaned_dir = Path(f'{path}/cleaned/{split}')
        cleaned_dir.mkdir(parents=True, exist_ok=True)

        for rec_df,name in zip(rec_dfs,FILENAMES[1:]):
            rec_df.to_csv(f'{cleaned_dir}/{name}.csv')
        patients.reset_index(inplace=True)
        patients.to_csv(f'{cleaned_dir}/patients.csv', index_label='indx')
        data_tables[1].to_csv(f'{cleaned_dir}/patient_demographics.csv')
        print(f'Saved cleaned "{split}" data to {cleaned_dir}')

        if split == 'train':
            codes_dir = Path(f'{cleaned_dir}/codes')
            codes_dir.mkdir(parents=True, exist_ok=True)
            for code_df,name in zip(code_tables, FILENAMES):
                code_df.to_csv(f'{codes_dir}/code_{name}.csv', index_label='indx')
            print(f'Saved vocab code tables to {codes_dir}')

# Cell
def load_cleaned_ehrdata(path):
    '''Load cleaned, age-filtered EHR data'''

    csv_names = FILENAMES.copy()
    csv_names.insert(1,'patient_demographics')

    train_dfs = [pd.read_csv(f'{path}/cleaned/train/{fname}.csv', low_memory=False, index_col=0) for fname in csv_names]
    valid_dfs = [pd.read_csv(f'{path}/cleaned/valid/{fname}.csv', low_memory=False, index_col=0) for fname in csv_names]
    test_dfs  = [pd.read_csv(f'{path}/cleaned/test/{fname}.csv', low_memory=False, index_col=0) for fname in csv_names]

    return train_dfs, valid_dfs, test_dfs

# Cell
def load_ehr_vocabcodes(path):
    '''Load codes for vocabs'''

    code_dfs = [pd.read_csv(f'{path}/cleaned/train/codes/code_{fname}.csv', low_memory=False, na_filter=False, index_col=0) for fname in FILENAMES]

    return code_dfs


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_device": "00_setup.ipynb",
         "read_settings": "00_setup.ipynb",
         "DEVICE": "00_setup.ipynb",
         "settings": "00_setup.ipynb",
         "DATA_STORE": "00_setup.ipynb",
         "LOG_STORE": "00_setup.ipynb",
         "MODEL_STORE": "00_setup.ipynb",
         "EXPERIMENT_STORE": "00_setup.ipynb",
         "PATH_1K": "00_setup.ipynb",
         "PATH_10K": "00_setup.ipynb",
         "PATH_20K": "00_setup.ipynb",
         "PATH_100K": "00_setup.ipynb",
         "FILENAMES": "00_setup.ipynb",
         "SYNTHEA_DATAGEN_DATES": "00_setup.ipynb",
         "CONDITIONS": "00_setup.ipynb",
         "LABELS": "00_setup.ipynb",
         "LOG_NUMERICALIZE_EXCEP": "00_setup.ipynb",
         "read_raw_ehrdata": "01_preprocessing_clean.ipynb",
         "split_patients": "01_preprocessing_clean.ipynb",
         "split_ehr_dataset": "01_preprocessing_clean.ipynb",
         "cleanup_pts": "01_preprocessing_clean.ipynb",
         "cleanup_obs": "01_preprocessing_clean.ipynb",
         "cleanup_algs": "01_preprocessing_clean.ipynb",
         "cleanup_crpls": "01_preprocessing_clean.ipynb",
         "cleanup_meds": "01_preprocessing_clean.ipynb",
         "cleanup_img": "01_preprocessing_clean.ipynb",
         "cleanup_procs": "01_preprocessing_clean.ipynb",
         "cleanup_cnds": "01_preprocessing_clean.ipynb",
         "cleanup_immns": "01_preprocessing_clean.ipynb",
         "cleanup_dataset": "01_preprocessing_clean.ipynb",
         "extract_ys": "01_preprocessing_clean.ipynb",
         "insert_age": "01_preprocessing_clean.ipynb",
         "clean_raw_ehrdata": "01_preprocessing_clean.ipynb",
         "load_cleaned_ehrdata": "01_preprocessing_clean.ipynb",
         "load_ehr_vocabcodes": "01_preprocessing_clean.ipynb",
         "EhrVocab": "02_preprocessing_vocab.ipynb",
         "ObsVocab": "02_preprocessing_vocab.ipynb",
         "EhrVocabList": "02_preprocessing_vocab.ipynb",
         "get_all_emb_dims": "02_preprocessing_vocab.ipynb",
         "collate_codes_offsts": "03_preprocessing_transform.ipynb",
         "get_codenums_offsts": "03_preprocessing_transform.ipynb",
         "get_demographics": "03_preprocessing_transform.ipynb",
         "Patient": "03_preprocessing_transform.ipynb",
         "get_pckl_dir": "03_preprocessing_transform.ipynb",
         "PatientList": "03_preprocessing_transform.ipynb",
         "cpu_cnt": "03_preprocessing_transform.ipynb",
         "create_all_ptlists": "03_preprocessing_transform.ipynb",
         "preprocess_ehr_dataset": "03_preprocessing_transform.ipynb",
         "EHRDataSplits": "04_data.ipynb",
         "LabelEHRData": "04_data.ipynb",
         "EHRDataset": "04_data.ipynb",
         "EHRData": "04_data.ipynb",
         "accuracy": "05_metrics.ipynb",
         "null_accuracy": "05_metrics.ipynb",
         "ROC": "05_metrics.ipynb",
         "MultiLabelROC": "05_metrics.ipynb",
         "plot_rocs": "05_metrics.ipynb",
         "plot_train_valid_rocs": "05_metrics.ipynb",
         "auroc_score": "05_metrics.ipynb",
         "auroc_ci": "05_metrics.ipynb",
         "save_to_checkpoint": "06_learn.ipynb",
         "load_from_checkpoint": "06_learn.ipynb",
         "get_loss_fn": "06_learn.ipynb",
         "RunHistory": "06_learn.ipynb",
         "train": "06_learn.ipynb",
         "evaluate": "06_learn.ipynb",
         "fit": "06_learn.ipynb",
         "predict": "06_learn.ipynb",
         "plot_loss": "06_learn.ipynb",
         "plot_losses": "06_learn.ipynb",
         "plot_aurocs": "06_learn.ipynb",
         "plot_train_valid_aurocs": "06_learn.ipynb",
         "plot_fit_results": "06_learn.ipynb",
         "summarize_prediction": "06_learn.ipynb",
         "count_parameters": "06_learn.ipynb",
         "dropout_mask": "07_models.ipynb",
         "InputDropout": "07_models.ipynb",
         "linear_layer": "07_models.ipynb",
         "create_linear_layers": "07_models.ipynb",
         "init_lstm": "07_models.ipynb",
         "EHR_LSTM": "07_models.ipynb",
         "init_cnn": "07_models.ipynb",
         "conv_layer": "07_models.ipynb",
         "EHR_CNN": "07_models.ipynb",
         "get_data": "08_experiment.ipynb",
         "get_optimizer": "08_experiment.ipynb",
         "get_model": "08_experiment.ipynb",
         "Experiment": "08_experiment.ipynb"}

modules = ["setup.py",
           "preprocessing/clean.py",
           "preprocessing/vocab.py",
           "preprocessing/transform.py",
           "data.py",
           "metrics.py",
           "learn.py",
           "models.py",
           "experiment.py"]

doc_url = "https://corazonlabs.github.io/lemonade/"

git_url = "https://github.com/corazonlabs/lemonade/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 03_preprocessing_transform.ipynb (unless otherwise specified).

__all__ = ['collate_codes_offsts', 'get_codenums_offsts', 'get_demographics', 'Patient', 'get_pckl_dir', 'PatientList',
           'cpu_cnt', 'create_all_ptlists', 'preprocess_ehr_dataset']

# Cell
from ..setup import *
from .clean import *
from .vocab import *
from fastai.imports import *
import torch.multiprocessing as multiprocessing

# Cell
def collate_codes_offsts(rec_df, age_start, age_stop, age_in_months=False):
    """Return a single patient's EmbeddingBag lookup codes and offsets for the given age span and age units"""
    codes  = []
    offsts = [0]
    age_span = age_stop - age_start
    if rec_df.empty:
        codes = ['xxnone'] * age_span
        offsts = list(range(age_span))
    else:
        for i in range(age_start, age_stop, 1):
            if age_in_months: res = (rec_df.code[rec_df.age_months == i]).values
            else            : res = (rec_df.code[rec_df.age == i]).values
            if len(res) > 0:
                codes.extend(res)
                if i < age_stop - 1: offsts.append(offsts[-1] + len(res))
            else:
                codes.append('xxnone')
                if i < age_stop - 1: offsts.append(offsts[-1] + 1)

    assert len(offsts) == age_span
    return codes, offsts

# Cell
def get_codenums_offsts(rec_dfs, all_vocabs, age_start, age_stop, age_in_months):
    '''Get numericalized record codes and offsets for a patient for a given age span'''
    all_codes_offsts = [collate_codes_offsts(df, age_start, age_stop, age_in_months) for df in rec_dfs]
    obs_vocab, alg_vocab, crpl_vocab, med_vocab, img_vocab, proc_vocab, cnd_vocab, imm_vocab = all_vocabs

    obs_c,  obs_o  = obs_vocab.numericalize (all_codes_offsts[0][0]), all_codes_offsts[0][1]
    alg_c,  alg_o  = alg_vocab.numericalize (all_codes_offsts[1][0]), all_codes_offsts[1][1]
    crpl_c, crpl_o = crpl_vocab.numericalize(all_codes_offsts[2][0]), all_codes_offsts[2][1]
    med_c,  med_o  = med_vocab.numericalize (all_codes_offsts[3][0]), all_codes_offsts[3][1]
    img_c,  img_o  = img_vocab.numericalize (all_codes_offsts[4][0]), all_codes_offsts[4][1]
    proc_c, proc_o = proc_vocab.numericalize(all_codes_offsts[5][0]), all_codes_offsts[5][1]
    cnd_c,  cnd_o  = cnd_vocab.numericalize (all_codes_offsts[6][0]), all_codes_offsts[6][1]
    imm_c,  imm_o  = imm_vocab.numericalize (all_codes_offsts[7][0]), all_codes_offsts[7][1]

    all_codenums = [obs_c,alg_c,crpl_c,med_c,img_c,proc_c,cnd_c,imm_c]
    all_offsts   = [obs_o,alg_o,crpl_o,med_o,img_o,proc_o,cnd_o,imm_o]

    return all_codenums, all_offsts

# Cell
def get_demographics(demograph_vector, demographics_vocabs, age_mean, age_std):
    '''Numericalize demographics and normalize age for a given patient'''
    bday, bmonth, byear, marital, race, ethnicity, gender, birthplace, city, state, zipcode = demographics_vocabs
    demograph_vector = demograph_vector.fillna('xxnone')
    demographics = []

    birthdate = pd.Timestamp(demograph_vector[0])

    demographics.extend(bday.numericalize      ([birthdate.day]))
    demographics.extend(bmonth.numericalize    ([birthdate.month]))
    demographics.extend(byear.numericalize     ([birthdate.year]))
    demographics.extend(marital.numericalize   ([demograph_vector[1]]))
    demographics.extend(race.numericalize      ([demograph_vector[2]]))
    demographics.extend(ethnicity.numericalize ([demograph_vector[3]]))
    demographics.extend(gender.numericalize    ([demograph_vector[4]]))
    demographics.extend(birthplace.numericalize([demograph_vector[5]]))
    demographics.extend(city.numericalize      ([demograph_vector[6]]))
    demographics.extend(state.numericalize     ([demograph_vector[7]]))
    demographics.extend(zipcode.numericalize   ([demograph_vector[8]]))
    age = (demograph_vector[9] - age_mean) / age_std

    return demographics, age

# Cell
class Patient():
    '''Class defining a patient object that holds all numericalized / transformed data for a single patient'''
    def __init__(self, nums, offsts, demographics, age_now, birthdate, diabetes, stroke, alzheimers, coronaryheart, ptid):
        self.obs_nums  = torch.tensor(nums[0])
        self.alg_nums  = torch.tensor(nums[1])
        self.crpl_nums = torch.tensor(nums[2])
        self.med_nums  = torch.tensor(nums[3])
        self.img_nums  = torch.tensor(nums[4])
        self.proc_nums = torch.tensor(nums[5])
        self.cnd_nums  = torch.tensor(nums[6])
        self.imm_nums  = torch.tensor(nums[7])

        self.obs_offsts  = torch.tensor(offsts[0])
        self.alg_offsts  = torch.tensor(offsts[1])
        self.crpl_offsts = torch.tensor(offsts[2])
        self.med_offsts  = torch.tensor(offsts[3])
        self.img_offsts  = torch.tensor(offsts[4])
        self.proc_offsts = torch.tensor(offsts[5])
        self.cnd_offsts  = torch.tensor(offsts[6])
        self.imm_offsts  = torch.tensor(offsts[7])

        self.demographics = torch.tensor(demographics)
        self.age_now      = torch.tensor([age_now])

        self.ptid = ptid
        self.birthdate = birthdate
        self.diabetes = diabetes
        self.stroke = stroke
        self.alzheimers = alzheimers
        self.coronaryheart = coronaryheart

    def __repr__(self):
        return f'ptid:{self.ptid}, birthdate:{self.birthdate}, diabetes:{self.diabetes}, device:{self.alg_nums.device}'

    @classmethod
    def create(cls, rec_dfs, demograph, vocablist, ptid, birthdate, diabetes, stroke, alzheimers, coronaryheart, age_start, age_stop, age_in_months):
        '''Lookup codes, numericalize and then create patient object - given a patient id'''
        codenums, offsts  = get_codenums_offsts(rec_dfs, vocablist.records_vocabs, age_start, age_stop, age_in_months)
        demographics, age_now = get_demographics(demograph, vocablist.demographics_vocabs, vocablist.age_mean, vocablist.age_std)
        return cls(codenums, offsts, demographics, age_now, birthdate, diabetes, stroke, alzheimers, coronaryheart, ptid)

    def pin_memory(self):
        '''Call `torch.Tensor.pin_memory` for (all tensors of) this patient object'''

        if not self.obs_nums.is_pinned():
            self.obs_nums  = self.obs_nums.pin_memory()
            self.alg_nums  = self.alg_nums.pin_memory()
            self.crpl_nums = self.crpl_nums.pin_memory()
            self.med_nums  = self.med_nums.pin_memory()
            self.img_nums  = self.img_nums.pin_memory()
            self.proc_nums = self.proc_nums.pin_memory()
            self.cnd_nums  = self.cnd_nums.pin_memory()
            self.imm_nums  = self.imm_nums.pin_memory()

            self.obs_offsts  = self.obs_offsts.pin_memory()
            self.alg_offsts  = self.alg_offsts.pin_memory()
            self.crpl_offsts = self.crpl_offsts.pin_memory()
            self.med_offsts  = self.med_offsts.pin_memory()
            self.img_offsts  = self.img_offsts.pin_memory()
            self.proc_offsts = self.proc_offsts.pin_memory()
            self.cnd_offsts  = self.cnd_offsts.pin_memory()
            self.imm_offsts  = self.imm_offsts.pin_memory()

            self.demographics = self.demographics.pin_memory()
            self.age_now      = self.age_now.pin_memory()

        return self

    def to_gpu(self, non_block=False):
        '''Puts (all tensors of) this patient object on GPU'''
        self.obs_nums  = self.obs_nums.to(DEVICE, non_blocking=non_block)
        self.alg_nums  = self.alg_nums.to(DEVICE, non_blocking=non_block)
        self.crpl_nums = self.crpl_nums.to(DEVICE, non_blocking=non_block)
        self.med_nums  = self.med_nums.to(DEVICE, non_blocking=non_block)
        self.img_nums  = self.img_nums.to(DEVICE, non_blocking=non_block)
        self.proc_nums = self.proc_nums.to(DEVICE, non_blocking=non_block)
        self.cnd_nums  = self.cnd_nums.to(DEVICE, non_blocking=non_block)
        self.imm_nums  = self.imm_nums.to(DEVICE, non_blocking=non_block)

        self.obs_offsts  = self.obs_offsts.to(DEVICE, non_blocking=non_block)
        self.alg_offsts  = self.alg_offsts.to(DEVICE, non_blocking=non_block)
        self.crpl_offsts = self.crpl_offsts.to(DEVICE, non_blocking=non_block)
        self.med_offsts  = self.med_offsts.to(DEVICE, non_blocking=non_block)
        self.img_offsts  = self.img_offsts.to(DEVICE, non_blocking=non_block)
        self.proc_offsts = self.proc_offsts.to(DEVICE, non_blocking=non_block)
        self.cnd_offsts  = self.cnd_offsts.to(DEVICE, non_blocking=non_block)
        self.imm_offsts  = self.imm_offsts.to(DEVICE, non_blocking=non_block)

        self.demographics = self.demographics.to(DEVICE, non_blocking=non_block)
        self.age_now      = self.age_now.to(DEVICE, non_blocking=non_block)

        return self

# Cell
def get_pckl_dir(path, split, age_start, age_stop, age_in_months):
    '''Util function to construct pickle dir name - for persisting transformed `PatientList`s'''
    dir_name = ''
    dir_name += 'months' if age_in_months else 'years'
    dir_name += f'_{age_start}_to_{age_stop}'
    pckl_dir = Path(f'{path}/processed/{dir_name}/{split}')
    return pckl_dir

# Cell
multiprocessing.set_sharing_strategy('file_system')
cpu_cnt = int(multiprocessing.cpu_count())

class PatientList():
    '''A class to hold a list of `Patient` objects'''
    def __init__(self, pts, path, split, age_start, age_stop, age_in_months):
        self.items     = pts
        self.base_path = path
        self.split     = split
        self.age_start = age_start
        self.age_stop  = age_stop
        self.age_type  = 'months' if age_in_months else 'years'

    def __len__(self): return len(self.items)
    def __iter__(self): return iter(self.items)
    def __getitem__(self, idx):
        if isinstance(idx, (int,slice)): return self.items[idx]
        if isinstance(idx[0],bool):
            assert len(idx)==len(self) # bool mask
            return [o for m,o in zip(idx,self.items) if m]
        return [self.items[i] for i in idx]
    def __repr__(self):
        res  = f'{self.__class__.__name__} ({len(self)} items)\n'
        res += f'base path:{self.base_path}; split:{self.split}; age span:{self.age_stop - self.age_start} {self.age_type}\n'
        res += f'age_start:{self.age_start}; age_stop:{self.age_stop}; age_type:{self.age_type}\n'
        for item in self.items[:10]:
            res += f'{item.__repr__()}\n'
        if len(self)>10: res = res[:-1]+ '...]'
        return res

    def _create_pts_chunk(indx_chnk, all_dfs, vocablist, pckl_dir, age_start, age_stop, age_in_months, verbose):
        '''Parallelized function to run on one core and transform a single chunk of patients and save'''

        pts = []
        for indx in indx_chnk:
            vals = all_dfs[0].iloc[indx].values
            ptid, birthdate = vals[0], vals[1]
            diabetes, stroke, alzheimers, coronaryheart = vals[2], vals[4], vals[6], vals[8]

            rec_dfs = []
            for rec_df in all_dfs[2:]:
                try:
                    rec_dfs.append(rec_df.loc[[ptid]])
                except KeyError:
                    rec_dfs.append(pd.DataFrame())

            demograph = all_dfs[1].loc[ptid]
            pts.append(Patient.create(rec_dfs, demograph, vocablist, ptid, birthdate, diabetes, stroke, alzheimers, coronaryheart, age_start, age_stop, age_in_months))

        with open(f'{pckl_dir}/patients_{indx_chnk[0]}_{indx_chnk[-1]}.ptlist', 'wb') as pckl_f:
            pickle.dump(pts,pckl_f)

        if verbose: print(f'{multiprocessing.current_process().name}-- completed {len(indx_chnk)} patients')
        return len(pts)

    @classmethod
    def create_save(cls, all_dfs, vocablist, pckl_dir, age_start, age_stop, age_in_months, verbose=False):
        '''Function to parellelize (based on available CPU cores), transformation for all patients in given dataset and save `PatientList` object'''
        pckl_dir.mkdir(parents=True, exist_ok=True)
        pts, indx_chnks = [], []

        total_pts = len(all_dfs[0])
        all_indxs = np.arange(total_pts)
        chnk_sz = total_pts // (cpu_cnt-1)
        for i in range(0, total_pts, chnk_sz):
            indx_chnks.append(list(all_indxs[i:i+chnk_sz]))

        pool = multiprocessing.Pool(processes=cpu_cnt)
        parallelize = partial(cls._create_pts_chunk, all_dfs=all_dfs, vocablist=vocablist, pckl_dir=pckl_dir, age_start=age_start, age_stop=age_stop, age_in_months=age_in_months, verbose=verbose)
        all_chunks = pool.map(parallelize, indx_chnks)
        pool.close()

        print(f'{sum(all_chunks)} total patients completed, saved patient list to {pckl_dir}')

    @classmethod
    def load(cls, path, split, age_start, age_stop, age_in_months):
        '''Load previously created `PatientList` object'''
        pckl_dir = get_pckl_dir(path, split, age_start, age_stop, age_in_months)
        if not pckl_dir.exists(): raise Exception(f'"{pckl_dir}" does not exist, run pre-processing to create that dataset first.')
        ptlist = []
        for file in Path(pckl_dir).glob("*.ptlist"):
            with open(file, 'rb') as infile:
                ptlist.extend(pickle.load(infile))

        return(cls(ptlist, path, split, age_start, age_stop, age_in_months))

# Cell
def create_all_ptlists(path:Path, age_start:int, age_stop:int, age_in_months:bool, vocab_path:Path=None, verbose:bool=False, delete_existing:bool=True):
    '''Create and save `PatientList`s for train, valid and test given dataset path'''
    if vocab_path is None: vocab_path = path
    all_dfs_splits = load_cleaned_ehrdata(path) #train_dfs, valid_dfs, test_dfs
    splits = ['train', 'valid', 'test']
    vocablist = EhrVocabList.load(vocab_path)

    for all_dfs, split in zip(all_dfs_splits, splits):
        pckl_dir = get_pckl_dir(path, split, age_start, age_stop, age_in_months)
        if delete_existing:
            for file in Path(pckl_dir).glob("*.ptlist"):
                file.unlink()
        PatientList.create_save(all_dfs, vocablist, pckl_dir, age_start, age_stop, age_in_months, verbose)

# Cell
def preprocess_ehr_dataset(path, today, valid_pct=0.2, test_pct=0.2, obs_vocab_buckets=5,
                           age_start=0, age_stop=20, age_in_months=False, vocab_path=None, from_raw_data=False):
    '''Util function to do all preprocessing - split & clean raw dataset, create vocab lists and create patient lists'''
    if from_raw_data:
        print('------------------- Splitting and cleaning raw dataset -------------------')
        clean_raw_ehrdata(path, valid_pct, test_pct, today)
        print('------------------- Creating vocab lists -------------------')
        EhrVocabList.create(path, num_buckets=obs_vocab_buckets).save()
    else:
        print('Since data is pre-cleaned, skipping Cleaning, Splitting and Vocab-creation')

    print('------------------- Creating patient lists -------------------')
    create_all_ptlists(path, age_start, age_stop, age_in_months, vocab_path)


# AUTOGENERATED! DO NOT EDIT! File to edit: 02_preprocessing_vocab.ipynb (unless otherwise specified).

__all__ = ['EhrVocab', 'ObsVocab', 'EhrVocabList', 'get_all_emb_dims']

# Cell
from ..setup import *
from .clean import *
from fastai.imports import *
from datetime import date

# Cell
class EhrVocab():
    '''Vocab class for most EHR datatypes'''
    def __init__(self, itoc, ctoi, ctod=None):
        self.itoc = itoc
        self.ctoi = ctoi
        if ctod is not None: self.ctod = ctod
        self.vocab_size = len(self.itoc)

    @classmethod
    def create(cls, codes_df):
        '''Create vocab object (itoc, ctoi and maybe ctod) from the codes df'''
        desc_exists = 'desc' in codes_df.columns
        codes_df = codes_df.astype({'code':'str'})
        itoc = list(codes_df.code.unique())  #old --> list(set(codes_df.code))
        itoc.insert(0,'xxnone')
        itoc.insert(1,'xxunk')

        ctoi = {code: i for i, code in enumerate(itoc)}

        if desc_exists:
            codes_df.set_index('code', inplace=True)
            ctod = {}
            ctod[itoc[0]] = "Nothing recorded"
            ctod[itoc[1]] = "Unknown"
            for code in itoc[2:]:
                ctod[code] = set(codes_df.loc[code].desc)

        return cls(itoc, ctoi, ctod) if desc_exists else cls(itoc, ctoi)

    def get_emb_dims(self, αd=0.5736):
        '''Get embedding dimensions'''
        return self.vocab_size, round(6* αd * (self.vocab_size**0.25))

    def numericalize(self, codes, log_excep=LOG_NUMERICALIZE_EXCEP, log_dir=LOG_STORE):
        '''Lookup and return indices for codes'''

        if log_excep:
            today = date.today().strftime("%Y-%m-%d")
            if not os.path.isdir(log_dir): os.mkdir(log_dir)
            logfile = f'{log_dir}/{today}_numericalize_exceptions.log'

        res = []
        try:
            res = [self.ctoi[str(code)] for code in codes] #no big performance benefit
        except KeyError:
            for code in codes:
                try:
                    res.append(self.ctoi[str(code)])
                except KeyError:
                    res.append(self.ctoi['xxunk'])
                    if log_excep:
                        with open(logfile, 'a') as log:
                            log.write(f'\ncode: {code}')

        return res

    def textify(self, indxs):
        '''Lookup and return descriptions for codes'''
        if hasattr(self, 'ctod'):
            res = [ (self.itoc[i], self.ctod[self.itoc[i]]) for i in indxs ]
        else:
            res = [ (self.itoc[i]) for i in indxs ]
        return res

# Cell
class ObsVocab (EhrVocab):
    '''Special Vocab class for Observation codes'''
    def __init__(self, vocab_df):
        self.vocab_df = vocab_df
        self.vocab_size = len(vocab_df)

    def numericalize(self, codes, log_excep=LOG_NUMERICALIZE_EXCEP, log_dir=LOG_STORE):
        '''Numericalize observation codes (return indices for codes)'''

        if log_excep:
            today = date.today().strftime("%Y-%m-%d")
            if not os.path.isdir(log_dir): os.mkdir(log_dir)
            logfile = f'{log_dir}/{today}_numericalize_exceptions.log'

        indxs = []
        for code in codes:
            if code in ['xxnone','xxunk']: indxs.extend(self.vocab_df[(self.vocab_df['code'] == code)].index.tolist())
            else:
                c,v,u,t = code.split('||')
                if t == 'numeric':
                    filt_df = self.vocab_df[(self.vocab_df['code'] == c) & (self.vocab_df['units'] == u) & (self.vocab_df['type'] == t)]
                    res = filt_df.iloc[(filt_df.value - float(v)).abs().argsort()[:1]].index.tolist()
                else:
                    res = self.vocab_df[(self.vocab_df['code'] == c) & (self.vocab_df['value'] == v) & \
                                               (self.vocab_df['units'] == u) & (self.vocab_df['type'] == t)].index.tolist()
                if len(res) == 0:
                    indxs.extend(self.vocab_df[(self.vocab_df['code'] == 'xxunk')].index.tolist())
                    if log_excep:
                        with open(logfile, 'a') as log:
                            log.write(f'\ncode in ObsVocab: {code}')
                else            : indxs.extend(res)

        assert len(codes) == len(indxs), "Possible bug, not all codes being numericalized"
        return indxs

    def textify(self, indxs):
        '''Textify observation codes (returns codes and descriptions)'''
        txts = []
        for i in indxs:
            c,d,v,u,t = self.vocab_df.iloc[i]
            if i == 0: txts.append((c, d))
            else:      txts.append((f'{c}||{v}||{u}||{t}', d))
        assert len(indxs) == len(txts), "Possible bug, not all indxs being textified"
        return txts

    @classmethod
    def create(cls, obs_codes, num_buckets=5):
        '''Create vocab object from observation codes'''
        numerics = pd.DataFrame(obs_codes.loc[obs_codes['type'] == 'numeric',:])
        texts = pd.DataFrame(obs_codes.loc[obs_codes['type'] == 'text',:])
        numerics = numerics.astype({'value':'float'}, copy=False)
        vocab_rows = []

        for code in numerics.orig_code.unique():
            this_code = numerics.loc[numerics['orig_code'] == code]
            for unit in this_code.units.unique():
                this_unit = this_code.loc[this_code['units'] == unit]
                for val in np.linspace(this_unit.value.min(), this_unit.value.max(), num=num_buckets):
                    vocab_rows.append([code,this_unit.desc.iloc[0],val,unit,'numeric'])

        for code in texts.orig_code.unique():
            this_code = texts.loc[texts['orig_code'] == code]
            for unit in this_code.units.unique():
                this_unit = this_code.loc[this_code['units'] == unit]
                for val in this_unit.value.unique():
                    vocab_rows.append([code,this_unit.desc.iloc[0],val,unit,'text'])

        vocab_rows.insert(0, ['xxnone','Nothing recorded','xxnone','xxnone','xxnone'])
        vocab_rows.insert(1, ['xxunk','Unknown','xxunk','xxunk','xxunk'])
        obs_vocab = pd.DataFrame(data=vocab_rows, columns=['code','desc','value','units','type'])
        assert obs_codes.orig_code.nunique() == obs_vocab.code.nunique()-2, "Possible bug, obs_code nuniques don't match"
        return cls(obs_vocab)

# Cell
class EhrVocabList:
    '''Class to create and hold all vocab objects for an entire dataset'''
    def __init__(self, demographics_vocabs, records_vocabs, age_mean, age_std, path):
        self.demographics_vocabs, self.records_vocabs, self.path = demographics_vocabs, records_vocabs, path
        self.age_mean, self.age_std = age_mean, age_std

    @classmethod
    def create(cls, path, num_buckets=5):
        '''Read all code dfs from the dataset path and create all vocab objects'''
        demographics_vocabs, records_vocabs = [], []
        code_dfs = load_ehr_vocabcodes(path)

        def _get_demographics_codes(pt_codes):
            code_dfs = []
            code_dfs.extend([pd.DataFrame(range(1, 32, 1), columns=['code'])]) #31 days
            code_dfs.extend([pd.DataFrame(range(1, 13, 1), columns=['code'])]) #12 months
            code_dfs.extend([pd.DataFrame(range(1900, pd.Timestamp.today().year + 1, 1), columns=['code'])]) #years 1900 to now
            code_dfs.extend([pd.DataFrame(pt_codes.marital.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.race.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.ethnicity.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.gender.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.birthplace.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.city.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.state.dropna().unique(), columns=['code'])])
            code_dfs.extend([pd.DataFrame(pt_codes.zip.dropna().unique(), columns=['code'])])
            age_mean, age_std = pt_codes.age_now_days.mean(), pt_codes.age_now_days.std()
            return code_dfs, age_mean, age_std

        demographics_codes, age_mean, age_std = _get_demographics_codes(code_dfs[0])
        demographics_vocabs.extend([EhrVocab.create(codes_df) for codes_df in demographics_codes])
        records_vocabs.extend([ObsVocab.create(code_dfs[1], num_buckets)])
        records_vocabs.extend([EhrVocab.create(codes_df) for codes_df in code_dfs[2:]])
        return cls(demographics_vocabs, records_vocabs, age_mean, age_std, path)

    def save(self):
        '''Save vocablist (containing all vocab objects for the dataset)'''
        pckl_dir = Path(f'{self.path}/processed')
        pckl_dir.mkdir(parents=True, exist_ok=True)
        pckl_f = open(f'{pckl_dir}/vocabs.vocablist', 'wb')
        pickle.dump(self, pckl_f)
        pckl_f.close()
        print(f'Saved vocab lists to {pckl_dir}')

    @classmethod
    def load(cls, path):
        '''Load previously created vocablist object (containing all vocab objects for the dataset)'''
        infile = open(f'{path}/processed/vocabs.vocablist','rb')
        ehrVocabList = pickle.load(infile)
        infile.close()
        return ehrVocabList

# Cell
def get_all_emb_dims(EhrVocabList, αd=0.5736):
    '''Get embedding dimensions for all vocab objects of the dataset'''
    demographics_dims = [vocab.get_emb_dims(αd) for vocab in EhrVocabList.demographics_vocabs]
    recs_dims          = [vocab.get_emb_dims(αd) for vocab in EhrVocabList.records_vocabs]

#     emb_dims_list = [vocab.get_emb_dims() for vocab in vocabs_list]
    demographics_dims_width = recs_dims_width = 0
    for emb_dim in demographics_dims:
        demographics_dims_width += emb_dim[1]
    for emb_dim in recs_dims:
        recs_dims_width += emb_dim[1]

    return demographics_dims, recs_dims, demographics_dims_width, recs_dims_width


# AUTOGENERATED! DO NOT EDIT! File to edit: 05_metrics.ipynb (unless otherwise specified).

__all__ = ['accuracy', 'null_accuracy', 'ROC', 'MultiLabelROC', 'plot_rocs', 'plot_train_valid_rocs', 'auroc_score',
           'auroc_ci']

# Cell
from .setup import *
from fastai.imports import *
from sklearn import metrics as skl_metrics, preprocessing as skl_preproc

# Cell
def accuracy(y:'y_true', yhat:'yhat_prob', threshold:float=0.5) -> float:
    '''Accuracy (percentage of correct predictions) for binary classification'''
    return ((yhat>threshold) == y).float().mean()

# Cell
def null_accuracy(y:'y_true') -> float:
    '''Null accuracy for binary classification: accuracy that could be achieved by always predicting the most frequent class'''
    null_mean = y.float().mean()
    return max(null_mean, 1-null_mean)

# Cell
class ROC:
    '''Class to hold Receiver Operating Characteristic (ROC) and AUROC (area under ROC curve) score for a single class'''
    def __init__(self, y, yhat):
        self.fpr, self.tpr, self.thresholds = skl_metrics.roc_curve(y, yhat)
        self.auroc = skl_metrics.roc_auc_score(y, yhat)

    def optimal_thresh(self):
        '''Calculate optimal threshold (on ROC curve) for a single class'''
        opt_idx = np.argmin(np.sqrt(np.square(1-self.tpr) + np.square(self.fpr)))
        return self.thresholds[opt_idx]

    def plot(self, label, title):
        '''Plot ROC curve for a single class'''
        plot_rocs([self], labels=[label], title=title)

# Cell
class MultiLabelROC:
    '''Class that holds `ROC` objects for multiple classes'''
    def __init__(self, y, yhat, labels):
        self.labels = labels
        self.ROCs = dict()
        for i in range(len(labels)):
            self.ROCs[labels[i]] = ROC(y[:, i], yhat[:, i])

    def plot(self, title):
        '''Plot multiple ROC curves in a single plot'''
        plot_rocs(self.ROCs, labels=self.labels, title=title, multilabel=True)

# Cell
def plot_rocs(ROCs, labels, title='ROC curve', multilabel=False, axis=None):
    '''Plot one (single-label) or multiple (multi-label) ROC curves'''
    if axis == None:
        fig = plt.figure(figsize=(8,5))
        axis = fig.add_axes([0,0,1,1])
    for i in range(len(ROCs)):
        if multilabel: axis.plot(ROCs[labels[i]].fpr, ROCs[labels[i]].tpr, label=f'{labels[i]} - {ROCs[labels[i]].auroc:0.3f}')
        else: axis.plot(ROCs[i].fpr, ROCs[i].tpr, label=f'{labels[i]} - {ROCs[i].auroc:0.3f}')
    axis.set_xlim([0.0, 1.0])
    axis.set_ylim([0.0, 1.0])
    axis.set_title(title)
    axis.set_xlabel('False Positive Rate (1 - Specificity)')
    axis.set_ylabel('True Positive Rate (Sensitivity)')
    axis.legend(loc="lower right")
    axis.grid(True)

# Cell
def plot_train_valid_rocs(train_ROC, valid_ROC, labels, multilabel=False):
    '''Convenience fn to plot train and valid ROC curves side by side'''
    if multilabel:
        fig, axes = plt.subplots(1,2, figsize=(15,5))
        plt.tight_layout()
        plot_rocs(train_ROC, labels, title='Train ROC curves', multilabel=True, axis=axes[0])
        plot_rocs(valid_ROC, labels, title='Valid ROC curves', multilabel=True, axis=axes[1])
    else:
        plot_rocs([train_ROC, valid_ROC], ['train', 'valid'], title='Train & Valid ROC Curves')

# Cell
def auroc_score(y, yhat, average=None):
    '''Return scikit_learn auroc score'''
    return skl_metrics.roc_auc_score(y, yhat, average=average)

# Cell
def auroc_ci(y, yhat):
    '''Returns 95% confidence interval for auroc'''
    n_bootstraps = 1000
    rng_seed = 42  # control reproducibility
    bootstrapped_scores = []

    rng = np.random.RandomState(rng_seed)
    for i in range(n_bootstraps):
        # bootstrap by sampling with replacement on the prediction indices
        indices = rng.randint(0, len(y), len(y))
        if len(np.unique(y[indices])) < 2: continue   # reject this sample
        score = skl_metrics.roc_auc_score(y[indices], yhat[indices])
        bootstrapped_scores.append(score)

    sorted_scores = np.array(bootstrapped_scores)
    sorted_scores.sort()

    confidence_lower = sorted_scores[int(0.025 * len(sorted_scores))]
    confidence_upper = sorted_scores[int(0.975 * len(sorted_scores))]
    return round(confidence_lower,3), round(confidence_upper,3)